// graphics_win.cpp
// Revision 7-jun-2006


#include "graphics_impl.h"


#ifdef BLASSIC_USE_WINDOWS_GRAPHICS


#include "util.h"
#include "key.h"
#include "error.h"
#include "showerror.h"
#include "trace.h"

#include <iostream>
#include <map>

#include <process.h>
#include <windows.h>
#undef min
#undef max

#include <cassert>
#define ASSERT assert


#if defined __CYGWIN32__ || defined __CYGWIN__
// This macros are from Anders Norlander, modified to add
// the cast to start_proc.
/* Macro uses args se we can cast proc to LPTHREAD_START_ROUTINE
   in order to avoid warings because of return type */
#define _beginthreadex(security, stack_size, start_proc, arg, flags, pid) \
CreateThread (security, stack_size, (LPTHREAD_START_ROUTINE) start_proc, \
	arg, flags, (LPDWORD) pid)
#define _endthreadex ExitThread
#endif

// Use Polyline to draw a point.
#define USE_POLY


using util::touch;
using util::dim_array;

using blassic::keynames::strCLICK;
using blassic::keynames::strSCLICK;
using blassic::keynames::strRELEASE;
using blassic::keynames::strSRELEASE;

using std::cerr;
using std::endl;
using std::map;
using std::make_pair;


namespace blassic {

namespace graphics_impl {


class CriticalSection {
public:
	CriticalSection ()
	{
		InitializeCriticalSection (& cs);
	}
	~CriticalSection ()
	{
		DeleteCriticalSection (& cs);
	}
	void enter ()
	{
		EnterCriticalSection (& cs);
	}
	void leave ()
	{
		LeaveCriticalSection (& cs);
	}
private:
	CRITICAL_SECTION cs;
};


class QueueKeyWin : public QueueKey {
private:
	void lock ();
	void unlock ();
	CriticalSection cs;
};


void QueueKeyWin::lock ()
{
	cs.enter ();
}

void QueueKeyWin::unlock ()
{
	cs.leave ();
}


class ImageMsWin : public Image {
public:
	ImageMsWin (int x1, int y1, int x2, int y2, HDC hdcPixmap);
	int getwidth () const;
	int getheight () const;
	HDC gethdc ();
private:
	HBITMAP img;
	HDC hdcImg;
	int width;
	int height;
};


ImageMsWin::ImageMsWin (int x1, int y1, int x2, int y2, HDC hdcPixmap)
{
	width= x2 - x1 + 1;
	height= y2 - y1 + 1;
	img= CreateCompatibleBitmap (hdcPixmap, width, height);
	if (img == NULL)
		throw ErrFunctionCall;
	hdcImg= CreateCompatibleDC (hdcPixmap);
	if (hdcImg == NULL)
	{
		DeleteObject (img);
		throw ErrFunctionCall;
	}
	HGDIOBJ hgdiobj= SelectObject (hdcImg, img);
	if (hgdiobj == NULL ||
		hgdiobj == reinterpret_cast <HGDIOBJ> (GDI_ERROR) )
	{
		DeleteDC (hdcImg);
		DeleteObject (img);
		throw ErrFunctionCall;
	}

	if (BitBlt (hdcImg, 0, 0, width, height, hdcPixmap, x1, y1, SRCCOPY)
		== 0)
	{
		DeleteDC (hdcImg);
		DeleteObject (img);
		throw ErrFunctionCall;
	}
}

int ImageMsWin::getwidth () const
{
	return width;
}

int ImageMsWin::getheight () const
{
	return height;
}

HDC ImageMsWin::gethdc ()
{
	return hdcImg;
}


class GraphicsSystemMsWin : public GraphicsSystem {
public:
	GraphicsSystemMsWin ();
	Screen * newScreen ();

	bool check_thread ();
	void create_thread ();
	void destroy_thread ();
	DWORD get_thread_id () const;
	ATOM get_atomClassMain () const;
	ATOM get_atomClassBorder () const;
	void wait_sync_event ();
	void set_sync_event ();
private:
	ATOM atomClassMain;
	ATOM atomClassBorder;
	HANDLE hthread;
	DWORD idthread;

	HANDLE hEvent;
};


class PainterMsWin;


class ScreenMsWin : public Screen {
public:
	ScreenMsWin (GraphicsSystemMsWin & system_n);
private:
	void raw_create (int width, int height, int rwidth, int rheight,
		RotateMode rotmode, InkNum paperink);
public:
	void set_title (const std::string & title);

	void set_synchro (bool synchro_n);
	bool get_synchro () const;

	LRESULT handle_WM_PAINT
		(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
	LRESULT handle_WM_PAINT_border
		(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
	LRESULT handle_border_proc
		(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
	LRESULT handle_window_proc
		(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);

	void checkwait ();
	void pushkey (const std::string & strkey);

	void init_pixmap ();

	void show ();

	void update_window (int x, int y, int width, int height);
	void update_window ();

	void setink (InkNum inknum, const ColorRGB & rgb);
	void active_foreground ();
	void active_background ();

	void setactiveink (InkNum inknum);

	void set_border (InkNum inknum);
	void set_border_width (int newwidth);

	void raw_plot (int x, int y);

	int raw_test (int x, int y);
	bool load_char_image (int x, int y, int xzoom, int yzoom,
		unsigned char (& ch) [8] );

	Image * raw_getImage (int x1, int y1, int x2, int y2);
	void raw_putImage (Image & image, int x, int y, int mode);

	Painter * raw_getPainter (InkNum inkpaint, InkNum inkborder,
		int minx, int miny, int maxx, int maxy);

	void real_plot_points (ShortPoint point [], int npoints);
	void real_plot_wide_points (ShortPoint point [], int npoints,
		int widex, int widey,
		int xmin, int ymin, int xmax, int ymax);

	void raw_draw_line (int prevx, int prevy, int x, int y);
	void create_in_thread (int width, int height);

	QueueKey & getQueueKey ();

	int raw_xmouse () const;
	int raw_ymouse () const;

	int keypressed (int keynum);

	void ring_my_bell ();
private:
	void raw_copy_rectangle (int x1from, int y1from,
		int x2from, int y2from, int xto, int yto);
	void raw_invert_rectangle (int x1, int y1, int x2, int y2);
	void raw_rectangle (int x1, int y1, int x2, int y2);
	void direct_fill_rectangle (int x1, int y1, int x2, int y2);
public:
	void destroy ();
	void wait_event ();
	void setdrawmode (int drawmodenum);

private:
	GraphicsSystemMsWin & system;

	int realwidth, realheight;
	RotateMode rotatemode;
	int drawmode;
	int border_width;

	bool synchro_mode;

	HWND hwndMain;
	HWND hwndBorder;
	HBITMAP pixmap;
	HDC hdc;
	HDC hdcPixmap;
	HBRUSH hbrushBorder;

	struct UsedWinColor {
		HPEN pen;
		ColorRGB rgb;
	};
	UsedWinColor defaultinks [PREDEFINED_INKS];
	typedef map <InkNum, UsedWinColor> usedinks_t;
	usedinks_t usedinks;

	UsedWinColor & mapcolor (InkNum color);
	UsedWinColor & mapnewcolor (InkNum color);

	bool border_set;
	InkNum border_ink;
	InkNum active_ink;

	void raw_init_colors ();

	static const int drawmode_copy= R2_COPYPEN;
	static const int drawmode_invert= R2_NOT;

	// Keyboard

	static const unsigned int MAXKEYCODE= 511;
	static const unsigned int NOUS= 0x1FF;
	static const unsigned int inkeytocode [MAXINKEYCODE + 1];
	std::vector <bool> keycode_pressed;

	QueueKeyWin queuekey;

	void keycode_press (unsigned int keycode);
	void keycode_release (unsigned int keycode);

	static int getbitbltmode (int mode);

	// Mouse pointer

	int xmousepos, ymousepos;

	// Main thread waiting for events.
	bool waiting_event;

	friend class PainterMsWin;
};


class PainterMsWin : public Painter {
public:
	PainterMsWin (ScreenMsWin & screen_n,
		InkNum inkpaint, InkNum inkborder,
		int minx, int miny, int maxx, int maxy);
	~PainterMsWin ();
private:
	ScreenMsWin & screen;
	COLORREF colorpaint, colorborder;
	int olddrawmode;
	bool oldsync;
	void plot (Point p);
	bool do_check (Point p);
	int x_min, x_max, y_min, y_max;
};

PainterMsWin::PainterMsWin (ScreenMsWin & screen_n,
		InkNum inkpaint, InkNum inkborder,
		int minx, int miny, int maxx, int maxy) :
	Painter (minx, miny, maxx, maxy),
	screen (screen_n)
{
	ScreenMsWin::UsedWinColor & ciup= screen.mapcolor (inkpaint);
	colorpaint= RGB (ciup.rgb.r, ciup.rgb.g, ciup.rgb.b);
	ScreenMsWin::UsedWinColor & ciub= screen.mapcolor (inkborder);
	colorborder= RGB (ciub.rgb.r, ciub.rgb.g, ciub.rgb.b);

	screen.setactiveink (inkpaint);
	olddrawmode= screen.drawmode;
	screen.setdrawmode (0);
	oldsync= screen.synchro_mode;
	screen.synchro_mode= true;
	x_min= y_min= 65000;
	x_max= y_max= -65000;
}

PainterMsWin::~PainterMsWin ()
{
	screen.synchro_mode= oldsync;
	screen.update_window (x_min, y_min,
		x_max - x_min + 1, y_max - y_min + 1);
}

void PainterMsWin::plot (Point p)
{
	int x= p.x, y= p.y;
	if (x_min > x)
		x_min= x;
	if (y_min > y)
		y_min= y;
	if (x_max < x)
		x_max= x;
	if (y_max < y)
		y_max= y;
	screen.raw_plot (x, y);
}

bool PainterMsWin::do_check (Point p)
{
	COLORREF cr= GetPixel (screen.hdcPixmap, p.x, p.y);
	return ! (cr == colorpaint || cr == colorborder);
}

static const UINT
	WM_USER_CREATE_WINDOW= WM_USER + 3,
	WM_USER_DESTROY_WINDOW= WM_USER + 4,
	WM_USER_WAITING_EVENT= WM_USER + 5;

struct ThreadParams {
	HANDLE hEvent;
	int width, height;
};

struct ThreadCreateWindowParams {
	ScreenMsWin & screen;
	int width, height;
	ThreadCreateWindowParams (ScreenMsWin & screen_n, int w, int h) :
		screen (screen_n),
		width (w),
		height (h)
	{ }
};

unsigned WINAPI threadproc (void * arg);

LRESULT APIENTRY windowproc_main
	(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT APIENTRY windowproc_border
	(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);


GraphicsSystemMsWin::GraphicsSystemMsWin () :
	atomClassMain (0),
	atomClassBorder (0),
	hthread (NULL),
	idthread (0)
{
	TRF;

	const bool showfailinfo= showdebuginfo ();

	WNDCLASS wndclass;
	wndclass.style= CS_NOCLOSE | CS_OWNDC;
	wndclass.lpfnWndProc= windowproc_main;
	wndclass.cbClsExtra= 0;
	wndclass.cbWndExtra= sizeof (ScreenMsWin *);
	wndclass.hInstance= GetModuleHandle (0);
	wndclass.hIcon= 0;
	wndclass.hCursor= LoadCursor (NULL, IDC_ARROW);
	wndclass.hbrBackground= HBRUSH (NULL);
	wndclass.lpszMenuName= 0;
	wndclass.lpszClassName= program_name.c_str ();
	atomClassMain= RegisterClass (& wndclass);
	if (atomClassMain == 0)
	{
		//if (showfailinfo)
		//{
			//cerr << "Error registering class" << endl;
			showlasterror ("Error registering class");
		//}
		throw ErrBlassicInternal;
	}
	TRMESSAGE (tr, "Window class registered");

	const std::string border_name= program_name + "_border";
	wndclass.style= CS_NOCLOSE;
	wndclass.lpfnWndProc= windowproc_border;
	wndclass.lpszClassName= border_name.c_str ();
	atomClassBorder= RegisterClass (& wndclass);
	if (atomClassBorder == 0)
	{
		if (showfailinfo)
			cerr << "Error registering border class" << endl;
		throw ErrBlassicInternal;
	}
	TRMESSAGE (tr, "Border class registered");

	// Create sync event.
	const BOOL auto_reset= FALSE;
	const BOOL initial_nonsignaled= FALSE;
	hEvent= CreateEvent (NULL, auto_reset, initial_nonsignaled, NULL);
	if (hEvent == 0)
	{
		showlasterror ("CreateEvent failed");
		throw ErrBlassicInternal;
	}
	else
	{
		TRMESSAGE (tr, "Sync event created");
	}
}

bool GraphicsSystemMsWin::check_thread ()
{
	return hthread != NULL;
}

void GraphicsSystemMsWin::create_thread ()
{
	TRF;

	ThreadParams tp= { hEvent, 0, 0 };
	hthread= HANDLE (_beginthreadex (NULL, 0, threadproc,
		& tp, 0, (unsigned int *) (& idthread) ) );
	if (hthread == NULL)
	{
		showlasterror ("Error creating graphics thread");
		throw ErrBlassicInternal;
	}
	TRMESSAGE (tr, "Thread created, waiting sync");
	wait_sync_event ();
}

void GraphicsSystemMsWin::destroy_thread ()
{
	if (idthread)
	{
		BOOL r= PostThreadMessage (idthread, WM_QUIT, 0, 0);
		if (r == 0)
		{
			TerminateThread (hthread, 0);
		}
		else
		{
			WaitForSingleObject (hthread, INFINITE);
		}
		CloseHandle (hthread);
		idthread= 0;
		hthread= NULL;
	}
}

DWORD GraphicsSystemMsWin::get_thread_id () const
{
	return idthread;
}

ATOM GraphicsSystemMsWin::get_atomClassMain () const
{
	return atomClassMain;
}

ATOM GraphicsSystemMsWin::get_atomClassBorder () const
{
	return atomClassBorder;
}

void GraphicsSystemMsWin::wait_sync_event ()
{
	DWORD r= WaitForSingleObject (hEvent, INFINITE);
	if (r !=  WAIT_OBJECT_0)
	{
		showlasterror ("Synchronization with graphics thread failed");
	}
}

void GraphicsSystemMsWin::set_sync_event ()
{
	SetEvent (hEvent);
}

Screen * GraphicsSystemMsWin::newScreen ()
{
	TRF;

	return new ScreenMsWin (* this);
}


class ProtectWindow {
	HWND hwnd;
public:
	ProtectWindow (HWND hwnd) : hwnd (hwnd) { }
	~ProtectWindow ()
	{
		if (hwnd)
			DestroyWindow (hwnd);
	}
	void release ()
	{
		hwnd= 0;
	}
};

class ProtectPixmap {
	HBITMAP & pixmap;
	bool active;
public:
	ProtectPixmap (HBITMAP & pixmap) : pixmap (pixmap), active (true) { }
	~ProtectPixmap ()
	{
		if (active)
		{
			DeleteObject (pixmap);
			pixmap= NULL;
		}
	}
	void release ()
	{
		active= false;
	}
};

void thread_create_window (ThreadCreateWindowParams & param)
{
	param.screen.create_in_thread (param.width, param.height);
}

unsigned WINAPI threadproc (void * arg)
{
	ThreadParams * ptp= reinterpret_cast <ThreadParams *> (arg);
	HANDLE hEvent= ptp->hEvent;

	//thread_create_window (tp->width, tp->height);

	// Ensure the message queue exist before the main thread continues.
	// Perhpas unnecesary with the new method, but...
	MSG msg;
	PeekMessage (& msg, NULL, 0, UINT (-1), PM_NOREMOVE);

	SetEvent (hEvent);

	//if (! window_created)
	//	return 0;

	while (GetMessage (& msg, NULL, 0, 0) )
	{
		switch (msg.message)
		{
		case WM_USER_CREATE_WINDOW:
			//thread_create_window (msg.wParam, msg.lParam);
			{
				ThreadCreateWindowParams * pthwparam=
					reinterpret_cast
						<ThreadCreateWindowParams *>
						(msg.lParam);
				thread_create_window (* pthwparam);
				SetEvent (hEvent);
			}
			break;
		#if 0
		case WM_USER_DESTROY_WINDOW:
			if (DestroyWindow (hwndMain) == 0)
				cerr << "Error destroying: " <<
					GetLastError () << endl;
			break;
		#endif
		default:
			TranslateMessage (& msg);
			DispatchMessage (& msg);
		}
	}
	return 0;
}

LRESULT APIENTRY windowproc_main
	(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_NCCREATE:
		{
			LPCREATESTRUCT lpcs=
				reinterpret_cast <LPCREATESTRUCT> (lParam);
			SetWindowLong (hwnd, GWL_USERDATA,
				reinterpret_cast <LONG> (lpcs->lpCreateParams) );
		}
		return TRUE;
	case WM_CREATE:
		return FALSE;
	case WM_GETMINMAXINFO:
		return DefWindowProc (hwnd, uMsg, wParam, lParam);
	default:
		ScreenMsWin * pscreen= reinterpret_cast <ScreenMsWin *>
			(GetWindowLong (hwnd, GWL_USERDATA) );
		ASSERT (pscreen);
		return pscreen->handle_window_proc
			(hwnd, uMsg, wParam, lParam);
	}
}

LRESULT APIENTRY windowproc_border
	(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_NCCREATE:
		{
			LPCREATESTRUCT lpcs=
				reinterpret_cast <LPCREATESTRUCT> (lParam);
			SetWindowLong (hwnd, GWL_USERDATA,
				reinterpret_cast <LONG> (lpcs->lpCreateParams) );
		}
		return TRUE;
	case WM_CREATE:
		return FALSE;
	case WM_GETMINMAXINFO:
		return DefWindowProc (hwnd, uMsg, wParam, lParam);

	default:
		ScreenMsWin * pscreen= reinterpret_cast <ScreenMsWin *>
			(GetWindowLong (hwnd, GWL_USERDATA) );
		ASSERT (pscreen);
		return pscreen->handle_border_proc
			(hwnd, uMsg, wParam, lParam);
	}
}

ScreenMsWin::ScreenMsWin (GraphicsSystemMsWin & system_n) :
	Screen (system_n),
	system (system_n),
	rotatemode (RotateNone),
	drawmode (drawmode_copy),
	//border_width (10),
	border_width (0),
	hwndMain (0),
	hwndBorder (0),
	pixmap (0),
	hdc (0),
	hdcPixmap (0),
	hbrushBorder (0),
	border_set (false),
	border_ink (static_cast <InkNum> (-1) ),
	active_ink (static_cast <InkNum> (-1) ),
	keycode_pressed (MAXKEYCODE + 1),
	xmousepos (-1),
	ymousepos (-1),
	waiting_event (false)
{
	for (size_t i= 0; i < PREDEFINED_INKS; ++i)
	{
		defaultinks [i].pen= 0;
		defaultinks [i].rgb.set (0, 0, 0);
	}
}

void ScreenMsWin::raw_create (int width, int height, int rwidth, int rheight,
	RotateMode rotmode, InkNum paperink)
{
	TRF;

	synchro_mode= false;

	#if 0
	if (hthread == NULL)
	{
		create_thread (width, height);
		if (hthread == NULL)
			throw ErrBlassicInternal;
	}
	#else

	if (! system.check_thread () )
		system.create_thread ();

	#endif

	touch (width, height);
	realwidth= rwidth;
	realheight= rheight;
	rotatemode= rotmode;
	if (! border_set)
		border_ink= paperink;
	active_ink= paperink;

	//BOOL r= PostThreadMessage (system.get_thread_id (),
	//	WM_USER_CREATE_WINDOW, WPARAM (width), LPARAM (height) );

	DWORD thread_id= system.get_thread_id ();
	if (thread_id == 0)
	{
		TRMESSAGE (tr, "Thread id not set");
		throw ErrBlassicInternal;
	}
	ThreadCreateWindowParams thwp (* this, rwidth, rheight);
	BOOL r= PostThreadMessage (thread_id,
		WM_USER_CREATE_WINDOW, WPARAM (0), LPARAM (& thwp) );

	if (r == 0)
	{
		static const char THREAD_COMM_ERROR []=
			"Communication with graphics thread failed";
		TRMESSAGE (tr, THREAD_COMM_ERROR);
		showlasterror (THREAD_COMM_ERROR);
		system.destroy_thread ();
		throw ErrBlassicInternal;
	}

	//WaitForSingleObject (hEvent, INFINITE);
	system.wait_sync_event ();

	set_title (default_title);

	// Deshabilitado temporalmente.
	#if 0

	if (! window_created)
	{
		if (showdebuginfo () )
			cerr << "Error creating window" << endl;
		throw ErrBlassicInternal;
	}
	#endif
}

void ScreenMsWin::init_pixmap ()
{
	TRF;

	RECT r = { 0, 0, realwidth, realheight };

	HPEN active= mapcolor (active_ink).pen;
	LOGPEN logpen;
	GetObject (active, sizeof (LOGPEN), & logpen);
	HBRUSH hbrush= CreateSolidBrush (logpen.lopnColor);
	FillRect (hdcPixmap, & r, hbrush);
	DeleteObject (hbrush);
}

void ScreenMsWin::show ()
{
	ShowWindow (hwndMain, SW_SHOWNORMAL);
	ShowWindow (hwndBorder, SW_SHOWNORMAL);
	SetForegroundWindow (hwndBorder);
}

void ScreenMsWin::update_window (int x, int y, int width, int height)
{
	TRF;

	BitBlt (hdc, x, y, width, height, hdcPixmap,
		x, y, SRCCOPY);
}

void ScreenMsWin::update_window ()
{
	#if 0
	TRF;

	BitBlt (hdc, 0, 0, realwidth, realheight, hdcPixmap,
		0, 0, SRCCOPY);
	#endif

	update_window (0, 0, realwidth, realheight);
}

void ScreenMsWin::setink (InkNum inknum, const ColorRGB & rgb)
{
	TRF;

	ASSERT (hdcPixmap);
	UsedWinColor & ciu= mapnewcolor (inknum);
	COLORREF newcolor=
		GetNearestColor (hdcPixmap, RGB (rgb.r, rgb.g, rgb.b) );
	ciu.rgb.r= GetRValue (newcolor);
	ciu.rgb.g= GetGValue (newcolor);
	ciu.rgb.b= GetBValue (newcolor);
	HPEN newpen= CreatePen (PS_SOLID, 1, newcolor);
	if (ciu.pen != NULL)
		DeleteObject (ciu.pen);
	ciu.pen= newpen;
}

void ScreenMsWin::active_foreground ()
{
	TRF;

	InkNum pennum= get_graphics_pen ();
	active_ink= pennum;
	HPEN active= mapcolor (pennum).pen;
	SelectObject (hdc, active);
	SelectObject (hdcPixmap, active);
}

void ScreenMsWin::active_background ()
{
	TRF;

	InkNum papernum= get_graphics_paper ();
	active_ink= papernum;
	HPEN active= mapcolor (papernum).pen;
	SelectObject (hdc, active);
	SelectObject (hdcPixmap, active);
}

void ScreenMsWin::setactiveink (InkNum inknum)
{
	active_ink= inknum;
	HPEN active= mapcolor (active_ink).pen;
	SelectObject (hdc, active);
	SelectObject (hdcPixmap, active);
}

void ScreenMsWin::set_border (InkNum inknum)
{
	border_ink= inknum;
	border_set= true;
	if (hwndBorder == NULL)
		return;
	if (hbrushBorder != 0)
	{
		DeleteObject (hbrushBorder);
		hbrushBorder= 0;
	}
	InvalidateRect (hwndBorder, NULL, TRUE);
}

void ScreenMsWin::set_border_width (int newwidth)
{
	border_width= newwidth;
	if (hwndBorder == NULL)
		return;
	SetWindowPos (hwndBorder, NULL, 0, 0,
		realwidth + GetSystemMetrics (SM_CXDLGFRAME) * 2 +
			border_width * 2,
		realheight + GetSystemMetrics (SM_CYDLGFRAME) * 2 +
			GetSystemMetrics (SM_CYCAPTION) +
			border_width * 2,
		SWP_NOACTIVATE | SWP_NOCOPYBITS | SWP_NOMOVE |
			SWP_NOOWNERZORDER |
			SWP_NOZORDER);
	SetWindowPos (hwndMain, NULL, border_width, border_width,
		0, 0,
		SWP_NOACTIVATE | SWP_NOCOPYBITS |
			SWP_NOOWNERZORDER |
			SWP_NOSIZE |
			SWP_NOZORDER);
}

void ScreenMsWin::raw_plot (int x, int y)
{
	POINT p [2]= { {x, y,}, {x + 1, y} };
	Polyline (hdcPixmap, p, 2);
	if (! synchro_mode)
		Polyline (hdc, p, 2);
}

int ScreenMsWin::raw_test (int x, int y)
{
	COLORREF color= GetPixel (hdcPixmap, x, y);
	for (int i= 0; i < 16; ++i)
	{
		const ColorRGB & c= mapcolor (i).rgb;
		if (RGB (c.r, c.g, c.b) == color)
			return i;
	}
	for (usedinks_t::iterator it= usedinks.begin ();
		it != usedinks.end (); ++it)
	{
		const ColorRGB & c= it->second.rgb;
		if (RGB (c.r, c.g, c.b) == color)
			return it->first;
	}
	return -1;
}

bool ScreenMsWin::load_char_image (int x, int y, int xzoom, int yzoom,
	unsigned char (& ch) [8] )
{
	COLORREF back= GetPixel (hdcPixmap, x, y);
	bool fFore= false;
	COLORREF fore= 0;
	for (int i= 0, ipos= y; i < 8; ++i, ipos+= yzoom)
		for (int j= 0, jpos= x; j < 8; ++j, jpos+= xzoom)
		{
			COLORREF c= GetPixel (hdcPixmap, jpos, ipos);
			bool bit;
			if (c == back)
				bit= false;
			else
			{
				if (! fFore)
				{
					fFore= true;
					fore= c;
				}
				if (c != fore)
					return false;
				bit= true;
			}
			if (xzoom > 1 || yzoom > 1)
			{
				const int iend= ipos + yzoom;
				const int jend= jpos + xzoom;
				for (int ii= ipos; ii < iend; ++ii)
					for (int jj= jpos; jj < jend; ++jj)
					{
						if (GetPixel (hdcPixmap,
								jj, ii) != c)
							return false;
					}
			}
			ch [i]<<= 1;
			ch [i]|= bit;
		}
	return true;
}

Image * ScreenMsWin::raw_getImage (int x1, int y1, int x2, int y2)
{
	TRF;

	return new ImageMsWin (x1, y1, x2, y2, hdcPixmap);
}

int ScreenMsWin::getbitbltmode (int mode)
{
	static const int bitbltmodesbynumber []= { SRCCOPY, SRCINVERT,
		SRCAND, SRCPAINT, DSTINVERT };

	if (mode < 0 ||
		size_t (mode) >= dim_array (bitbltmodesbynumber) )
	{
		throw ErrFunctionCall;
	}
	return bitbltmodesbynumber [mode];
}

void ScreenMsWin::raw_putImage (Image & image, int x, int y, int mode)
{
	TRF;

	ImageMsWin * pimagew= dynamic_cast <ImageMsWin *> (& image);
	if (pimagew == NULL)
		throw ErrFunctionCall;

	int maxwidth= realwidth - x + 1;
	int maxheight= realheight - y + 1;
	int w= std::min (pimagew->getwidth (), maxwidth);
	int h= std::min (pimagew->getheight (), maxheight);

	int modeused= getbitbltmode (mode);

	HDC hdcImg= pimagew->gethdc ();
	if (BitBlt (hdcPixmap, x, y, w, h, hdcImg, 0, 0, modeused) == 0)
		throw ErrFunctionCall;

	if (! synchro_mode)
		update_window (x, y, w, h);
}

Painter * ScreenMsWin::raw_getPainter (InkNum inkpaint, InkNum inkborder,
	int minx, int miny, int maxx, int maxy)
{
	return new PainterMsWin (* this, inkpaint, inkborder, minx, miny, maxx, maxy);
}

// Length of each segment: all are 2 points.
DWORD poly_points [64];
bool init_poly_points ()
{
	std::fill_n (poly_points, util::dim_array (poly_points), 2);
	return true;
}
bool poly_points_inited= init_poly_points ();


void ScreenMsWin::real_plot_points (ShortPoint point [], int npoints)
{
	TRF;

	// The PolyPolyline is a bit faster than doing each
	// point separately.
	// 64 * 2 points by segment
	static POINT p [64 * 2];

	for (int i= 0; i < npoints; ++i)
	{
		int xpos= point [i].x;
		int ypos= point [i].y;
		p [2 * i].x= xpos;
		p [2 * i].y= ypos;
		++xpos;

		p [2 * i + 1].x= xpos;
		p [2 * i + 1].y= ypos;
	}
	PolyPolyline (hdcPixmap, p, poly_points, npoints);
	if (! synchro_mode)
		PolyPolyline (hdc, p, poly_points, npoints);
}

void ScreenMsWin::real_plot_wide_points (ShortPoint point [], int npoints,
		int widex, int widey,
		int xmin, int ymin, int xmax, int ymax)
{
	TRF;

	if (widex == 1 || widey == 1)
	{
		if (widex == 1)
		{
			widex= 0;
			//--widey;
		}
		else
		{
			widey= 0;
			//--widex;
		}

		TRMESSAGE (tr, "Using PolyPolyline");
		// The PolyPolyline is a bit faster than doing each
		// point separately.
		// 64 * 2 points by segment
		static POINT p [64 * 2];

		for (int i= 0; i < npoints; ++i)
		{
			int xpos= point [i].x;
			int ypos= point [i].y;
			p [2 * i].x= xpos;
			p [2 * i].y= ypos;
			xpos+= widex;
			ypos+= widey;

			p [2 * i + 1].x= xpos;
			p [2 * i + 1].y= ypos;
		}
		PolyPolyline (hdcPixmap, p, poly_points, npoints);
		if (! synchro_mode)
			PolyPolyline (hdc, p, poly_points, npoints);
	}
	else
	{
		TRMESSAGE (tr, "Using FillRect");
		HPEN active= mapcolor (active_ink).pen;
		LOGPEN logpen;
		GetObject (active, sizeof (LOGPEN), & logpen);
		HBRUSH hbrush= CreateSolidBrush (logpen.lopnColor);
		for (int i= 0; i < npoints; ++i)
		{
			int x= point [i].x;
			int y= point [i].y;
			int x1= std::max (x, xmin);
			int y1= std::max (y, ymin);
			int x2= std::min (x + widex - 1, xmax) + 1;
			int y2= std::min (y + widey - 1, ymax) + 1;
			RECT r= { x1, y1, x2, y2 };
			FillRect (hdcPixmap, & r, hbrush);
			if (! synchro_mode)
				FillRect (hdc, & r, hbrush);
		}
		DeleteObject (hbrush);
	}
}

void ScreenMsWin::raw_draw_line (int prevx, int prevy, int x, int y)
{
	MoveToEx (hdcPixmap, prevx, prevy, 0);
	LineTo (hdcPixmap, x, y);
	LineTo (hdcPixmap, x + 1, y);
	if (! synchro_mode)
	{
		MoveToEx (hdc, prevx, prevy, 0);
		LineTo (hdc, x, y);
		LineTo (hdc, x + 1, y);
	}
}

ScreenMsWin::UsedWinColor & ScreenMsWin::mapcolor (InkNum color)
{
	if (color < PREDEFINED_INKS)
		return defaultinks [color];

	usedinks_t::iterator it= usedinks.find (color);
	if (it != usedinks.end () )
		return it->second;
	else
		return defaultinks [0];
}

ScreenMsWin::UsedWinColor & ScreenMsWin::mapnewcolor (InkNum color)
{
	if (color < PREDEFINED_INKS)
		return defaultinks [color];

	usedinks_t::iterator it= usedinks.find (color);
	if (it != usedinks.end () )
		return it->second;
	else
	{
		UsedWinColor n;
		return usedinks.insert (make_pair (color, n) ).first->second;
	}
}

void ScreenMsWin::raw_init_colors ()
{
	TRF;
	//active_foreground ();
}

void ScreenMsWin::raw_copy_rectangle (int x1from, int y1from,
	int x2from, int y2from, int xto, int yto)
{
	int w= x2from - x1from + 1;
	int h= y2from - y1from + 1;

	BitBlt (hdcPixmap, xto, yto, w, h,
		hdcPixmap, x1from, y1from, SRCCOPY);
	if (! synchro_mode)
		BitBlt (hdc, xto, yto, w, h,
			hdcPixmap, xto, yto, SRCCOPY);
}


void ScreenMsWin::raw_invert_rectangle (int x1, int y1, int x2, int y2)
{
	HBRUSH hbrush= (HBRUSH) GetStockObject (BLACK_BRUSH);
	HDC ahdc [2]= { hdc, hdcPixmap };
	for (size_t i= 0; i < 2; ++i)
	{
		HDC hdc= ahdc [i];
		SetROP2 (hdc, drawmode_invert);
		HBRUSH hold= (HBRUSH) SelectObject (hdc, hbrush);
		//Rectangle (hdc, x1, y1 + 6, x1 + 8, y1 + 8);
		Rectangle (hdc, x1, y1, x2, y2);
		SelectObject (hdc, hold);
		SetROP2 (hdc, drawmode);
	}
}

void ScreenMsWin::setdrawmode (int drawmodenum)
{
	static const int drawmodesbynumber []= {
		R2_COPYPEN,
		R2_XORPEN,
		// Revisar los valores para and y or.
		R2_MASKPEN,
		R2_MERGEPEN,
		R2_NOT
	};

	if (drawmodenum < 0 || drawmodenum > drawmodenum_max)
		throw ErrFunctionCall;
	drawmode= drawmodesbynumber [drawmodenum];
	SetROP2 (hdcPixmap, drawmode);
	SetROP2 (hdc, drawmode);
}

void ScreenMsWin::create_in_thread (int width, int height)
{
	TRF;

	hwndBorder= CreateWindow (
		LPCTSTR (system.get_atomClassBorder () ),
		//default_title.c_str (),
		NULL,
		WS_SYSMENU | WS_MINIMIZEBOX | WS_CLIPCHILDREN,
		0, 0,
		width + GetSystemMetrics (SM_CXDLGFRAME) * 2 +
			border_width * 2,
		height + GetSystemMetrics (SM_CYDLGFRAME) * 2 +
			GetSystemMetrics (SM_CYCAPTION) +
			border_width * 2,
		NULL,
		NULL,
		GetModuleHandle (0),
		static_cast <LPVOID> (this) );

	hwndMain= CreateWindow (
		LPCTSTR (system.get_atomClassMain () ),
		NULL,
		///*WS_VISIBLE | */ WS_SYSMENU | WS_MINIMIZEBOX,
		WS_CHILD,
		border_width, border_width,
		//width + GetSystemMetrics (SM_CXDLGFRAME) * 2,
		//height + GetSystemMetrics (SM_CYDLGFRAME) * 2 +
		//	GetSystemMetrics (SM_CYCAPTION),
		width, height,
		hwndBorder,
		NULL,
		GetModuleHandle (0),
		static_cast <LPVOID> (this) );
	if (hwndMain)
	{
		ProtectWindow pw (hwndMain);
		hdc= GetDC (hwndMain);
		if (hdc == NULL)
			return;
		pixmap= CreateCompatibleBitmap (hdc, width, height);
		if (pixmap == NULL)
			return;
		ProtectPixmap pp (pixmap);
		hdcPixmap= CreateCompatibleDC (hdc);
		if (hdcPixmap == NULL)
			return;
		SelectObject (hdcPixmap, pixmap);

		init_colors ();

		init_pixmap ();

		// Deshabilitado temporalmente.
		//window_created= true;

		ShowWindow (hwndMain, SW_SHOWNORMAL);
		//SetActiveWindow (hwndMain);
		SetForegroundWindow (hwndMain);
		pw.release ();
		pp.release ();
	}
}

void ScreenMsWin::destroy ()
{
	TRF;

	DestroyWindow (hwndBorder);
}

void ScreenMsWin::set_title (const std::string & title)
{
	//SetWindowText (hwndMain, title.c_str () );
	SetWindowText (hwndBorder, title.c_str () );
}


void ScreenMsWin::set_synchro (bool synchro_n)
{
	synchro_mode= synchro_n;
	if (! synchro_mode)
		update_window ();
}

bool ScreenMsWin::get_synchro () const
{
	return synchro_mode;
}

QueueKey & ScreenMsWin::getQueueKey ()
{
	return queuekey;
}

const unsigned int ScreenMsWin::inkeytocode [MAXINKEYCODE + 1]={
	// The key symbols indicated are the corresponding
	// to a spanish keyboard.
	0x008, //  0 Up
	0x00D, //  1 Right
	0x068, //  2 Down
	0x049, //  3 Numeric 9
	0x04d, //  4 Numeric 6
	0x051, //  5 Numeric 3
	0x11C, //  6 Intro
	0x05B, //  7 Numeric .
	0x064, //  8 Left
	0x000, //  9 Copy
	0x047, // 10 Numeric 7
	0x048, // 11 Numeric 8
	0x04C, // 12 Numeric 5
	0x04F, // 13 Numeric 1
	0x050, // 14 Numeric 2
	0x052, // 15 Numeric 0
	0x153, // 16 Delete
	0x01B, // 17 +
	0x01C, // 18 Return
	0x02B, // 19
	0x04B, // 20 Numeric 4
	0x02A, // 21 Shift
	0x056, // 22 \ on the CPC, <>
	0x01D, // 23 Control
	0x00D, // 24 
	0x00C, // 25 '
	0x01A, // 26 `[
	0x019, // 27 P
	0x028, // 28 '{
	0x027, // 29
	0x035, // 30 -
	0x034, // 31 .
	0x00B, // 32 0
	0x00A, // 33 9
	0x018, // 34 O
	0x017, // 35 I
	0x026, // 36 L
	0x025, // 37 K
	0x032, // 38 M
	0x033, // 39 ,
	0x009, // 40 8
	0x008, // 41 7
	0x016, // 42 U
	0x015, // 43 Y
	0x023, // 44 H
	0x024, // 45 J
	0x031, // 46 N
	0x039, // 47 space
	0x007, // 48 6
	0x006, // 49 5
	0x013, // 50 R
	0x014, // 51 T
	0x022, // 52 G
	0x021, // 53 F
	0x030, // 54 B
	0x02F, // 55 V
	0x005, // 56 4
	0x004, // 57 3
	0x012, // 58 E
	0x011, // 59 W
	0x01F, // 60 S
	0x020, // 61 D
	0x02E, // 62 C
	0x02D, // 63 X
	0x002, // 64 1
	0x003, // 65 2
	0x001, // 66 Escape
	0x010, // 67 Q
	0x00F, // 68 Tab
	0x01E, // 69 A
	0x03A, // 70 Caps lock
	0x02C, // 71 Z
	NOUS,  // 72 joystick on cpc, unasigned
	NOUS,  // 73 joystick on cpc, unasigned
	NOUS,  // 74 joystick on cpc, unasigned
	NOUS,  // 75 joystick on cpc, unasigned
	NOUS,  // 76 joystick on cpc, unasigned
	NOUS,  // 77 joystick on cpc, unasigned
	NOUS,  // 78 Inexistent
	0x00E, // 79 Backspace
};

void ScreenMsWin::keycode_press (unsigned int keycode)
{
	//std::cerr << "Pressed " << std::hex << std::setfill ('0') <<
	//	std::setw (3) << keycode << std::endl;
	if (keycode <= MAXKEYCODE)
		keycode_pressed [keycode]= true;
}

void ScreenMsWin::keycode_release (unsigned int keycode)
{
	//std::cerr << "Released " << std::hex << std::setfill ('0') <<
	//	std::setw (3) << keycode << std::endl;
	if (keycode <= MAXKEYCODE)
		keycode_pressed [keycode]= false;
}

int ScreenMsWin::keypressed (int keynum)
{
	if (keynum < 0 || keynum > static_cast <int> (MAXINKEYCODE) )
		return -1;

	const int shiftpressed= 32, ctrlpressed= 128;

	const unsigned int
		shift_left= 0x2A, shift_right= 0x36,
		control_left= 0x1D, control_right= 0x11D;

	bool shifted= keycode_pressed [shift_left] ||
		keycode_pressed [shift_right];
	bool controled= keycode_pressed [control_left] ||
		keycode_pressed [control_right];

	bool pressed;
	switch (keynum) {
	case 21:
		pressed= shifted;
		break;
	case 23:
		pressed= controled;
		break;
	default:
		pressed= keycode_pressed [inkeytocode [keynum] ];
	}
	if (! pressed)
		return -1;
	int r= 0;
	if (shifted)
		r|= shiftpressed;
	if (controled)
		r|= ctrlpressed;

	return r;
}

int ScreenMsWin::raw_xmouse () const
{
	return xmousepos;
}

int ScreenMsWin::raw_ymouse () const
{
	return ymousepos;
}

void ScreenMsWin::ring_my_bell ()
{
	MessageBeep (MB_ICONEXCLAMATION);
}

inline unsigned int getkeycode (LPARAM lParam)
{
	return (lParam & 0x001FF0000) >> 16;
}


LRESULT ScreenMsWin::handle_WM_PAINT
	(HWND hwnd, UINT /*uMsg*/, WPARAM /*wParam*/, LPARAM /*lParam*/)
{
	//err << "WM_PAINT " << width << ", " << height << endl;
	PAINTSTRUCT ps;
	/*HDC hdc= */ BeginPaint (hwnd, & ps);
	//BitBlt (hdc, 0, 0, realwidth, realheight,
	//	hdcPixmap, 0, 0, SRCCOPY);
	const RECT & r= ps.rcPaint;
	update_window (r.left, r.top,
		r.right - r.left, r.bottom - r.top);
	EndPaint (hwnd, & ps);
	return FALSE;
}

LRESULT ScreenMsWin::handle_WM_PAINT_border
	(HWND hwnd, UINT /*uMsg*/, WPARAM /*wParam*/, LPARAM /*lParam*/)
{
	if (hbrushBorder == 0)
	{
		HPEN active= mapcolor (border_ink).pen;
		LOGPEN logpen;
		GetObject (active, sizeof (LOGPEN), & logpen);
		hbrushBorder= CreateSolidBrush (logpen.lopnColor);
	}
	PAINTSTRUCT ps;
	HDC hdc= BeginPaint (hwnd, & ps);
	if (ps.fErase)
	{
		FillRect (hdc, & ps.rcPaint, hbrushBorder);
	}
	return FALSE;
}

LRESULT ScreenMsWin::handle_border_proc
	(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_PAINT:
		return handle_WM_PAINT_border (hwnd, uMsg, wParam, lParam);
	case WM_SETFOCUS:
		SetFocus (hwndMain);
		return FALSE;
	default:
		return DefWindowProc (hwnd, uMsg, wParam, lParam);
	}
}

LRESULT ScreenMsWin::handle_window_proc
	(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	LRESULT r;
	switch (uMsg)
	{
	case WM_SIZE:
		// A check here may be useful.
		r= TRUE;
		break;
	case WM_ERASEBKGND:
		r= TRUE;
		break;
	case WM_PAINT:
		r= handle_WM_PAINT (hwnd, uMsg, wParam, lParam);
		break;
	case WM_KEYDOWN:
	case WM_SYSKEYDOWN:
		keycode_press (getkeycode (lParam) );
		{
			WORD k= (WORD) wParam;
			//std::string str= string_from_virtual_key (k);
			switch (rotatemode)
			{
			case RotateNone:
				// Nothing to do.
				break;
			case Rotate90:
				switch (k)
				{
				case VK_LEFT:
					k= VK_UP; break;
				case VK_UP:
					k= VK_RIGHT; break;
				case VK_RIGHT:
					k= VK_DOWN; break;
				case VK_DOWN:
					k= VK_LEFT; break;
				}
				break;
			}
			std::string str= string_from_key_Win (k);
			if (! str.empty () )
			{
				pushkey (str);
				r= TRUE;
			}
			else
			{
				checkwait ();
				r= FALSE;
			}
		}
		break;
	case WM_KEYUP:
		keycode_release (getkeycode (lParam) );
		checkwait ();
		r= FALSE;
		break;
//	case WM_SYSKEYDOWN:
//		keycode_press (getkeycode (lParam) );
//		return FALSE;
	case WM_SYSKEYUP:
		keycode_release (getkeycode (lParam) );
		checkwait ();
		r= FALSE;
		break;
	case WM_CHAR:
		{
			char c= (char) wParam;
			pushkey (std::string (1, c) );
		}
		r= TRUE;
		break;
	case WM_MOUSEMOVE:
		xmousepos= LOWORD (lParam);
		ymousepos= HIWORD (lParam);
		checkwait ();
		//do_unrotate (xmousepos, ymousepos);
		r= TRUE;
		break;
	case WM_LBUTTONDOWN:
		pushkey (strCLICK);
		r= TRUE;
		break;
	case WM_RBUTTONDOWN:
		pushkey (strSCLICK);
		r= TRUE;
		break;
	case WM_LBUTTONUP:
		pushkey (strRELEASE);
		r= TRUE;
		break;
	case WM_RBUTTONUP:
		pushkey (strSRELEASE);
		r= TRUE;
		break;
	case WM_DESTROY:
		//SetEvent (hEvent);
		system.set_sync_event ();
		r= TRUE;
		break;
	case WM_USER_WAITING_EVENT:
		waiting_event= true;
		r= TRUE;
		break;
	default:
		r= DefWindowProc (hwnd, uMsg, wParam, lParam);
	}
	return r;
}

void ScreenMsWin::checkwait ()
{
	if (waiting_event)
	{
		system.set_sync_event ();
		waiting_event= false;
	}
}

void ScreenMsWin::pushkey (const std::string & strkey)
{
	queuekey.push (strkey);
	checkwait ();
}


void ScreenMsWin::raw_rectangle (int x1, int y1, int x2, int y2)
{
	HBRUSH hold= (HBRUSH) SelectObject (hdcPixmap,
		GetStockObject (NULL_BRUSH) );
	Rectangle (hdcPixmap, x1, y1, x2 + 1, y2 + 1);
	SelectObject (hdc, hold);
	if (! synchro_mode)
	{
		hold= (HBRUSH) SelectObject (hdc,
			GetStockObject (NULL_BRUSH) );
		Rectangle (hdc, x1, y1, x2 + 1, y2 + 1);
		SelectObject (hdc, hold);
	}
}

void ScreenMsWin::direct_fill_rectangle (int x1, int y1, int x2, int y2)
{
	RECT r = { x1, y1, x2 + 1, y2 + 1 };
	HPEN active= mapcolor (active_ink).pen;

	LOGPEN logpen;
	GetObject (active, sizeof (LOGPEN), & logpen);
	HBRUSH hbrush= CreateSolidBrush (logpen.lopnColor);
	FillRect (hdcPixmap, & r, hbrush);
	if (! synchro_mode)
		FillRect (hdc, & r, hbrush);
	DeleteObject (hbrush);
}

void ScreenMsWin::wait_event ()
{
	// Do nothing
	//Sleep (1000);

	SendMessage (hwndMain, WM_USER_WAITING_EVENT, 0, 0);
	system.wait_sync_event ();
}


GraphicsSystem * createSystemMsWin ()
{
	TRF;

	return new GraphicsSystemMsWin ();
}


} // namespace graphics

} // namespace blassic


#endif

// End of graphics_win.cpp
