// key.cpp
// Revision 25-apr-2006


#include "key.h"

#include "trace.h"


#include <map>


#include "config_win.h"


#if BLASSIC_USE_WINDOWS

#include <windows.h>

#endif


namespace blassic {


const std::string
	keynames::strPAGEUP   ("PAGEUP"),
	keynames::strPAGEDOWN ("PAGEDOWN"),
	keynames::strEND      ("END"),
	keynames::strHOME     ("HOME"),
	keynames::strLEFT     ("LEFT"),
	keynames::strUP       ("UP"),
	keynames::strRIGHT    ("RIGHT"),
	keynames::strDOWN     ("DOWN"),
	keynames::strINSERT   ("INSERT"),
	keynames::strDELETE   ("DELETE"),
	keynames::strF1       ("F1"),
	keynames::strF2       ("F2"),
	keynames::strF3       ("F3"),
	keynames::strF4       ("F4"),
	keynames::strF5       ("F5"),
	keynames::strF6       ("F6"),
	keynames::strF7       ("F7"),
	keynames::strF8       ("F8"),
	keynames::strF9       ("F9"),
	keynames::strF10      ("F10"),
	keynames::strF11      ("F11"),
	keynames::strF12      ("F12"),
	keynames::strENTER    ("\n"),
	keynames::strCLICK    ("CLICK"),
	keynames::strSCLICK   ("SCLICK"),
	keynames::strRELEASE  ("RELEASE"),
	keynames::strSRELEASE ("SRELEASE");


} // namespace blassic


#if BLASSIC_USE_WINDOWS


namespace {


typedef std::map <int, std::string> mapkeyWin_t;

mapkeyWin_t mapkeyWin;

inline void assignmapkeyWin (int code, const std::string & name)
{
	#ifndef NDEBUG
	mapkeyWin_t::iterator it;
	if ( (it= mapkeyWin.find (code) ) != mapkeyWin.end () )
	{
		std::cerr << "Error: key " << name <<
			" has equal code than " << it->second <<
			std::endl;
		throw 1;
	}
	#endif
	mapkeyWin [code]= name;
}

class InitMapkeyWin {
	InitMapkeyWin ();
	static InitMapkeyWin instance;
};

InitMapkeyWin InitMapkeyWin::instance;

InitMapkeyWin::InitMapkeyWin ()
{
	TRF;

	using namespace blassic::keynames;

	assignmapkeyWin (VK_PRIOR,  strPAGEUP);
	assignmapkeyWin (VK_NEXT,   strPAGEDOWN);
	assignmapkeyWin (VK_END,    strEND);
	assignmapkeyWin (VK_HOME,   strHOME);
	assignmapkeyWin (VK_LEFT,   strLEFT);
	assignmapkeyWin (VK_UP,     strUP);
	assignmapkeyWin (VK_RIGHT,  strRIGHT);
	assignmapkeyWin (VK_DOWN,   strDOWN);
	assignmapkeyWin (VK_INSERT, strINSERT);
	assignmapkeyWin (VK_DELETE, strDELETE);
	assignmapkeyWin (VK_F1,     strF1);
	assignmapkeyWin (VK_F2,     strF2);
	assignmapkeyWin (VK_F3,     strF3);
	assignmapkeyWin (VK_F4,     strF4);
	assignmapkeyWin (VK_F5,     strF5);
	assignmapkeyWin (VK_F6,     strF6);
	assignmapkeyWin (VK_F7,     strF7);
	assignmapkeyWin (VK_F8,     strF8);
	assignmapkeyWin (VK_F9,     strF9);
	assignmapkeyWin (VK_F10,    strF10);
	assignmapkeyWin (VK_F11,    strF11);
	assignmapkeyWin (VK_F12,    strF12);
}

} // namespace


std::string blassic::string_from_key_Win (int key)
{
	mapkeyWin_t::const_iterator it= mapkeyWin.find (key);
	if (it != mapkeyWin.end () )
		return it->second;
	return std::string ();
}


#endif


#if 0

namespace {

typedef std::map <unsigned int, std::string> mapkey_t;

mapkey_t mapkey;

inline void assignmapkey (unsigned int code, const std::string & name)
{
	#ifndef NDEBUG
	mapkey_t::iterator it;
	if ( (it= mapkey.find (code) ) != mapkey.end () )
	{
		std::cerr << "Error: key " << name <<
			" has equal code than " << it->second <<
			std::endl;
		throw 1;
	}
	#endif
	mapkey [code]= name;
}

bool initmapkey ()
{
	TRACEFUNC (tr, "initmapkey");

	#if BLASSIC_USE_WINDOWS

	assignmapkey (VK_PRIOR,  strPAGEUP);
	assignmapkey (VK_NEXT,   strPAGEDOWN);
	assignmapkey (VK_END,    strEND);
	assignmapkey (VK_HOME,   strHOME);
	assignmapkey (VK_LEFT,   strLEFT);
	assignmapkey (VK_UP,     strUP);
	assignmapkey (VK_RIGHT,  strRIGHT);
	assignmapkey (VK_DOWN,   strDOWN);
	assignmapkey (VK_INSERT, strINSERT);
	assignmapkey (VK_DELETE, strDELETE);
	assignmapkey (VK_F1,     strF1);
	assignmapkey (VK_F2,     strF2);
	assignmapkey (VK_F3,     strF3);
	assignmapkey (VK_F4,     strF4);
	assignmapkey (VK_F5,     strF5);
	assignmapkey (VK_F6,     strF6);
	assignmapkey (VK_F7,     strF7);
	assignmapkey (VK_F8,     strF8);
	assignmapkey (VK_F9,     strF9);
	assignmapkey (VK_F10,    strF10);
	assignmapkey (VK_F11,    strF11);
	assignmapkey (VK_F12,    strF12);

	#elif BLASSIC_USE_X

	assignmapkey (XK_Prior,     strPAGEUP);
	assignmapkey (XK_KP_Prior,  strPAGEUP);
	assignmapkey (XK_Next,      strPAGEDOWN);
	assignmapkey (XK_KP_Next,   strPAGEDOWN);
	assignmapkey (XK_End,       strEND);
	assignmapkey (XK_KP_End,    strEND);
	assignmapkey (XK_Home,      strHOME);
	assignmapkey (XK_KP_Home,   strHOME);
	assignmapkey (XK_Left,      strLEFT);
	assignmapkey (XK_KP_Left,   strLEFT);
	assignmapkey (XK_Up,        strUP);
	assignmapkey (XK_KP_Up,     strUP);
	assignmapkey (XK_Right,     strRIGHT);
	assignmapkey (XK_KP_Right,  strRIGHT);
	assignmapkey (XK_Down,      strDOWN);
	assignmapkey (XK_KP_Down,   strDOWN);
	assignmapkey (XK_Insert,    strINSERT);
	assignmapkey (XK_KP_Insert, strINSERT);
	assignmapkey (XK_Delete,    strDELETE);
	assignmapkey (XK_KP_Delete, strDELETE);
	assignmapkey (XK_F1,        strF1);
	assignmapkey (XK_F2,        strF2);
	assignmapkey (XK_F3,        strF3);
	assignmapkey (XK_F4,        strF4);
	assignmapkey (XK_F5,        strF5);
	assignmapkey (XK_F6,        strF6);
	assignmapkey (XK_F7,        strF7);
	assignmapkey (XK_F8,        strF8);
	assignmapkey (XK_F9,        strF9);
	assignmapkey (XK_F10,       strF10);
	assignmapkey (XK_F11,       strF11);
	assignmapkey (XK_F12,       strF12);

	#endif

	return true;
}

const bool mapkeyinited= initmapkey ();

} // namespace

std::string string_from_key (unsigned int key)
{
	mapkey_t::const_iterator it= mapkey.find (key);
	if (it != mapkey.end () )
		return it->second;
	return std::string ();
}

#endif

// End of key.cpp
